"""The CLI commands"""
import locale
import random
from typing import Tuple

import click

from tidier.core import Path, find_sub_files


def main(
        input_path: Path,
        output_path: Path,
        regex_match: str,
        regex_replace: str,
        exclude_patterns: Tuple,
        should_copy: bool,
        locale_code: str,
        all_files: bool,
        jalali_date: bool,
        log: bool,
        silent: bool = False,
) -> None:
    """Custom full option command for organizing files"""
    locale.setlocale(locale.LC_ALL, locale_code)

    exclude_patterns = list(exclude_patterns)
    if not all_files:
        exclude_patterns.append("**/.*")

    files = find_sub_files(input_path, regex_match, exclude_patterns)

    failed_files = []
    changed_paths = []
    with click.progressbar(files, item_show_func=lambda x: x.path.__str__() if x else "Done!") as files_bar:
        for file in files_bar:
            try:
                old_path = file.path
                formatted_path = file.formatter(regex_replace, jalali_date)
                new_path = file.rename(regex_match, formatted_path, output_path, should_copy)
                changed_paths.append((old_path, new_path))
            except Exception as e:
                failed_files.append((file, e))

    if failed_files:
        with open(Path.cwd() / "tidier_fails.txt", "w") as f:
            for file, error in failed_files:
                f.write(f"{file.path}\n")
        click.secho(f"[!] Failed files are saved to {Path.cwd() / 'tidier_fails.txt'}", err=True, fg="yellow")

    if log:
        for old_path, new_path in changed_paths:
            click.secho(f"{'Copied' if should_copy else 'Moved'} {old_path} to {new_path}", fg="green")
        for file, error in failed_files:
            click.secho(f"[!] Failed {'copying' if should_copy else 'moving'} {file.path} due: {error}", err=True,
                        fg="red")


@click.group()
def cli():
    """Tidier CLI"""
    pass


@cli.command()
@click.argument(
    "input_path",
    type=click.Path(exists=True, path_type=Path, file_okay=False),
)
@click.option(
    "-o",
    "--output",
    "output_path",
    help="Output directory",
    default=None,
    type=click.Path(
        file_okay=False,
        writable=True,
        path_type=Path
    ))
@click.option(
    "-m",
    "--match",
    "regex_match",
    help="Regex match",
    type=str,
    default="^(.+)$",
    show_default=True,
)
@click.option(
    "-r",
    "--replace",
    "regex_replace",
    help="Regex replace",
    type=str,
    default=r"%Y/%B/%d/{name}",
    show_default=True,
)
@click.option(
    "-e",
    "--exclude",
    "exclude_patterns",
    help="Excluding patterns",
    default=[],
    multiple=True,
)
@click.option(
    "-c",
    "--copy",
    "should_copy",
    help="Copy files instead of Move",
    default=False,
    show_default=True,
    is_flag=True,
)
@click.option(
    "-l",
    "--locale",
    "locale_code",
    help="Date and time locale",
    default="en_US",
    show_default=True,
)
@click.option(
    "-a",
    "--all",
    "all_files",
    help="Apply on all files including HIDDEN files",
    default=False,
    show_default=True,
    is_flag=True,
)
@click.option(
    "-j",
    "--jalali",
    "jalali_date",
    help="Use Jalali calendar",
    default=False,
    show_default=True,
    is_flag=True,
)
@click.option(
    "--log",
    help="Full log for changed paths",
    default=False,
    is_flag=True,
)
def tidy(
        input_path: Path,
        output_path: Path,
        regex_match: str,
        regex_replace: str,
        exclude_patterns: Tuple[str],
        should_copy: bool,
        locale_code: str,
        all_files: bool,
        jalali_date: bool,
        log: bool,
) -> None:
    """Custom full option command for organizing files"""
    main(input_path, output_path, regex_match, regex_replace, exclude_patterns, should_copy, locale_code, all_files, jalali_date, log)


@cli.command("tvshow")
@click.argument(
    "input_path",
    type=click.Path(exists=True, path_type=Path, file_okay=False),
)
@click.option(
    "-o",
    "--output",
    "output_path",
    help="Output directory",
    default=None,
    type=click.Path(
        file_okay=False,
        writable=True,
        path_type=Path
    ))
@click.option(
    "--ignore-subtitles",
    help="Ignore subtitles",
    default=False,
    show_default=True,
    is_flag=True,
)
@click.option(
    "-c",
    "--copy",
    "should_copy",
    help="Copy files instead of Move",
    default=False,
    show_default=True,
    is_flag=True,
)
@click.option(
    "--log",
    help="Full log for changed paths",
    default=False,
    is_flag=True,
)
def tv_show(
        input_path: Path,
        output_path: Path,
        ignore_subtitles: bool,
        should_copy: bool,
        log: bool
) -> None:
    """Auto Tidy TV Shows Episodes"""
    click.secho("[+] Searching for TV Shows", fg="green")
    main(
        input_path,
        output_path,
        r"^.*S0*(?P<season>\d{1,2})E0*(?P<episode>\d{1,2}).*(?P<ext>mkv|mp4)$",
        r"Season \g<season>/Episode \g<episode>.\g<ext>",
        (),
        should_copy,
        "en_US",
        False,
        False,
        log
    )
    if not ignore_subtitles:
        click.secho("[+] Searching subtitles...", fg="green")
        main(
            input_path,
            output_path,
            r"^.*S0*(?P<season>\d{1,2})E0*(?P<episode>\d{1,2})(?P<)*(?P<ext>srt|ssa)$",
            r"Season \g<season> | Subtitles/Episode \g<episode>.\g<ext>",
            (),
            should_copy,
            "en_US",
            False,
            False,
            log
        )


if __name__ == "__main__":
    cli()



"""File Class"""
import datetime
import mimetypes
import re
import shutil
from pathlib import Path

import jdatetime


class File:
    """File object with given path"""

    ALLOWED_FORMATTERS = ["name", "stem", "ext", "type", "suffix", "path", "parent"]

    def __init__(self, path: Path) -> None:
        """Description

        Args:
            path (Path) : Desired file's path.

        Returns:
            None

        Raises:
            ValueError : Given path is not a file.
        """
        if not path.is_file():
            raise ValueError
        self.path = path

    @property
    def parent(self) -> Path:
        """Finding parent of the file using `path`.

        Returns:
            File's parent
        """
        return self.path.parent

    @property
    def name(self) -> str:
        """Finding name of the file using `path`.

        Returns:
            File's name
        """
        return self.path.name

    @property
    def stem(self) -> str:
        """Finding name without extension of the file using `path`.

        Returns:
            File's Just name
        """
        return self.path.stem

    @property
    def suffix(self) -> str:
        """Finding suffix of the file using `path`.

        Returns:
            File's suffix
        """
        return self.path.suffix

    @property
    def ext(self) -> str:
        """Finding extension of the file using file's suffix.

        Returns:
            File's Extension
        """
        suffix = self.suffix
        if len(suffix) > 0:
            output = suffix[1:]
        else:
            output = "unknown"
        return output

    @property
    def type(self) -> str:
        """Guess the type of the file using `mimetypes`.

        Returns:
            File's type
        """
        guess_type = mimetypes.guess_type(self.name)[0]
        if guess_type:
            guess_type = guess_type.split("/")[0]
            return guess_type
        return "unknown"

    @property
    def modified_date(self) -> datetime.datetime:
        """Finding modified date & time.

        Returns:
            File's Modified DateTime
        """
        mtime = datetime.datetime.fromtimestamp(self.path.stat().st_mtime)
        return mtime

    @property
    def formatters_dict(self) -> dict:
        """Getting dictionary of file's attributes than can be use in formatting

        Returns:
            Dictionary of file's attributes
        """
        dic = {}
        for formatter in self.ALLOWED_FORMATTERS:
            dic[formatter] = getattr(self, formatter)
        return dic

    def formatter(self, string: str, jalali_date: bool) -> str:
        """Getting dictionary of file's attributes than can be use in formatting

        Args:
            string (str) : String to be formatted
            jalali_date (bool) : Flag to use `Jalali Calendar` or not

        Returns:
            Formatted string
        """
        if jalali_date:
            date = jdatetime.datetime.fromgregorian(datetime=self.modified_date)
        else:
            date = self.modified_date
        string = date.strftime(string)
        return string.format(**self.formatters_dict)

    def match(self, regex: str) -> bool:
        """Checking if file's name matches given regex.

        Args:
            regex (str) : Regex to be matched

        Returns:
            True if file's name matches given regex, False otherwise.
        """
        if regex is None:
            return True
        return bool(re.search(regex, self.name))

    def rename(self, regex: str, replacement: str, parent: Path = None, copy: bool = True) -> Path:
        """Rename file's name using given regex and replacement.

        Args:
            regex (str) : Regex to be matched
            replacement (str) : Replacement string
            parent (Path) : Parent directory of the new file
            copy (bool) : Flag to copy file or move

        Returns:
            Renamed file's path
        """
        new_path = Path(re.sub(regex, replacement, self.name))
        new_name = new_path.name
        if not parent:
            parent = self.path.parent
        new_parent_path = parent / new_path.parent
        new_parent_path.mkdir(parents=True, exist_ok=True)
        if copy:
            return shutil.copy(self.path, new_parent_path / new_name)
        else:
            return self.path.rename(new_parent_path / new_name)

    def __str__(self) -> str:
        """String representation of the file.

        Returns:
            File's path representation
        """
        return self.path.__str__()



"""Helper utilities"""
from typing import List, Iterable

from .classes import Path, File


def find_sub_files(dir_path: Path, match_regex: str = None, exclude_patterns: Iterable[str] = None) -> List[File]:
    """Find all files in given path matching with regex and exclude some

    Args:
        dir_path (Path) : Directory path to search files in
        exclude_patterns (Iterable[str]) : Patterns to exclude in finding files
        match_regex (str, optional): Search regex to filter files. Defaults to None.

    Return:
        List of founded Files object
    """

    all_sub_paths = dir_path.glob("**/*")
    excluded_paths: List[str] = []
    for pattern in exclude_patterns:
        excluded_paths += dir_path.glob(pattern)
    paths = [
        path
        for path in all_sub_paths
        if path not in excluded_paths
    ]
    files = [
        File(path)
        for path in paths
        if path.is_file()
    ]
    return list(filter(lambda file: file.match(match_regex), files))
